use std::io::prelude::*;
use std::net::{SocketAddr, TcpStream, ToSocketAddrs};
use std::time::{Duration, Instant};
use std::thread;
use std::io::BufReader;
use std::sync::mpsc;

use crate::results::*;


///Scan the ports on address by attempting a tcp connection.
pub fn tcp_scan(address: String, timeout: Duration, sleep: Duration, ports: Vec<u32>) -> std::io::Result<()> {
    println!("Scanning {}", address);

    for i in ports {
        let socket = resolve_add(&address.as_str(), &i);
        match TcpStream::connect_timeout(&socket, timeout) {
            Ok(_) => {
                println!("TCP/{} => ACCEPT", i);
                //s.write(b"Some padding")?;
            }
            Err(_) => {
                println!("TCP/{} => REJECT", i);
            }
        }

        thread::sleep(sleep.clone());
        //stream.write(b"Some great padding")?;
        //stream.read(&mut [0; 128])?;
    }
    Ok(())
}

///Takes either an hostname or an IP address and port, and resolve to a socket address.
///DNS recordings can lead to multiple entries : we take the first one.
fn resolve_add(host: &str, port: &u32) -> SocketAddr {
    let details = format!("{}:{}", host, port);
    let server: Vec<_> = details
        .to_socket_addrs()
        .expect("Can't resolve your address")
        .collect();
    return server[0];
}



///Try to stress the remote machine by attempting to every given port at once and sending payload
pub fn tcp_stress(payload: String, address: String, timeout: Duration, interval: Duration, ports: Vec<u32>, threads: u32) -> std::io::Result<()> {
    let mut handles = vec![];
    let (tx, rx) = mpsc::channel();

    for port in ports {
        let mut threads_left = threads;
        let tx2 = tx.clone();

        loop {
            let add = address.clone();
            let pay = payload.clone();
            let int = interval;
            let end_time = Instant::now() + int;
            let tx3 = tx2.clone();
            //pay.push_str("\r\n");


            handles.push(thread::spawn( move || {
                let socket = resolve_add(&add.as_str(), &port);
                loop {
                    //Check if it is time to stop
                    if Instant::now() >= end_time { break; }

                    let b4 = Instant::now();
                    match TcpStream::connect_timeout(&socket, timeout) {
                        Ok(mut s) => {
                            //Remote port answers
                            //Writing
                            s.write_all(pay.as_bytes()).unwrap();
                            s.flush().unwrap();

                            //Reading
                            s.set_read_timeout(Some(Duration::from_millis(500))).unwrap();
                            let mut reader = BufReader::new(s.try_clone().unwrap());
                            let mut answer = String::new();
                            let count = reader.read_line(&mut answer).unwrap_or(0);
                            if count > 0 {
                                println!("Read {} bytes for port {} -> {}", count, &port, answer);
                            }

                            let time = Instant::now() - b4;
                            let myres = StressResult { time: time, ok: true, rtype: ResType::Tcp(port.clone()) };
                            let _ = tx3.send(myres);

                            //Sleeping
                            thread::sleep(Duration::from_millis(10));

                        },
                        Err(_) => {
                            let time = Instant::now() - b4;
                            let myres = StressResult { time: time, ok: false, rtype: ResType::Tcp(port.clone()) };
                            let _ = tx3.send(myres);
                        }
                    }
                }
            }));
            *&mut threads_left -=1;
            if threads_left == 0 { break; }
        }
    }
    drop(tx);
    println!("Stressing {} using {} concurrent threads per port for {:?} \n", address, threads, interval);

    for handle in handles {
        let _ = handle.join();
    }

    let mut res: Vec<StressResult> = vec![];
    for r in rx.try_iter() {
        res.push(r);
    }
    let listofres = ListOfResults { results: res, restype: ResType::Tcp(1) };
    listofres.show_info();
    

    Ok(())
}


